/*
 Quasar E1 interface board Linux driver.

 Copyright (C) 2006-2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2011 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/ioctl.h>
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/device.h>
#if defined(HAVE_UNLOCKED_IOCTL) && defined(CONFIG_BKL)                                                                                                         
# include <linux/smp_lock.h> /* lock_kernel */
#endif
#include <linux/version.h>

#include "kal.h"
#include "version.h"
#include "pmc_e1_conf.h"
#include "pmc_e1.h"
#include "e12zap.h"
#include "pmc_e1_if.h"		/* part of ecfg, include after <linux/ioctl.h> */

/* Defines */
#define DRV_NAME "quasar"
#define WIN_SIZE 0x10000	/* PCI window size */

/* Global variables */
static dev_t dev_file_glob;	/* descriptop for devfs and sysfs */
static unsigned char device_number = 0;
static unsigned int cacheline = 0x20;
static unsigned int cacheline_bridge = 0x20;
static unsigned int latency = 0;
static int irqflags = IRQF_SHARED | IRQF_DISABLED;
LIST_HEAD(dev_list_head);

/* Global initialized structs */
static struct pci_device_id ptmc_e1_tbl[] = {
	{0x1172, 0x2222, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0,}
};

MODULE_DEVICE_TABLE(pci, ptmc_e1_tbl);

/* Function prototypes */
/* */

#ifdef PMC_E1_UDEV
static struct class *pmc_e1_class = NULL;
#else
#define ptmc_sysfs_create(a)
#define ptmc_sysfs_delete(a)
#endif				/* PMC_E1_UDEV */

/*========== Common functions ==========*/
static void init_ptmc_regs(const struct quasar_pmc *pmc)
{
	unsigned int output_2048 = 0;
	unsigned int use_2048_for_sync = 0;
	if (PMC_OLD_FIRMWARE != (pmc->bar0[0x0D] & 0xffff)) {
		output_2048 = PMC_CT_2048_OUT;	/* Set to output 2048 kb instead of 8 kb */
		/*    use_2048_for_sync = PMC_USE_FOR_SYNC_2048; */
	}
        /* Configure board */
	pmc->bar0[0x02] = 0xc1000000 | MODE1_MA_SL_SEL_HP | PMC_CT_TI_EN \
		| PMC_DMARST | PMC_CONS_DEL | output_2048 | use_2048_for_sync;
	udelay(1000);
	/* Wake up framers, hm magic reset */
	pmc->bar0[0x02] = 0x41000000 | PMC_WORKING_MODE | MODE1_FLOCK \
		| MODE1_MA_SL_SEL_HP | PMC_CT_TI_EN | PMC_DMARST \
		| PMC_CONS_DEL | output_2048 | use_2048_for_sync;
	udelay(1000);
	pmc->bar0[0x01] = 0x80000000;	/* Enable interrupts, in general */
	pmc->bar0[0x06] = 0x000000ff;	/* BS/LS: allow to sync to all E1 channels */
	pmc->bar0[0x04] = 0x000000ff;	/* TX_enable: enable transmit */
}


static void init_switch_map(const struct quasar_pmc *pmc)
{
	int i, j;

	/*
	   PC=>E1
	   DMA_LINE_2 => CT_BUS_4 => E1_LINE0
	   DMA_LINE_3 => CT_BUS_5 => E1_LINE1 */

	for (i = 0; i < 128; i++) {	/* LINE2 -> CT_BUS4 */
		pmc->bar0[0x2400 + i] = E1_TO_CTB_TABLE(2, i);
		pmc->bar0[0x2500 + i] = E1_TO_CTB_TABLE(3, i);
	}
	for (i = 0; i < 32; i++) {
		/* CT_BUS => E1; according to suggestion 
		   4 CT channel
		   CT_BUS4 -> LINE0 */
		pmc->bar0[0x1000 + i] = CTB_TO_E1_TABLE(4, i);
		pmc->bar0[0x1100 + i] = CTB_TO_E1_TABLE(4, (i + 32));
		pmc->bar0[0x1200 + i] = CTB_TO_E1_TABLE(4, (i + 64));
		pmc->bar0[0x1300 + i] = CTB_TO_E1_TABLE(4, (i + 96));
		/* CT_BUS5 -> LINE1 */
		pmc->bar0[0x1400 + i] = CTB_TO_E1_TABLE(5, i);
		pmc->bar0[0x1500 + i] = CTB_TO_E1_TABLE(5, (i + 32));
		pmc->bar0[0x1600 + i] = CTB_TO_E1_TABLE(5, (i + 64));
		pmc->bar0[0x1700 + i] = CTB_TO_E1_TABLE(5, (i + 96));
	}
	/* E1=>PC */
	for (i = 0; i < 32; i++) {
		/* E1 => CT_BUS */
		for (j = 0; j < 4; j++) {
			/* E1_LINE0 => CT_BUS_6 => WRITE_BUF_0 */
			pmc->bar0[0x2600 + i + 32 * j] =
			    E1_TO_CTB_TABLE(0, (i * 4 + j));
			/* E1_LINE1 => CT_BUS_7 => WRITE_BUF_1 */
			pmc->bar0[0x2700 + i + 32 * j] =
			    E1_TO_CTB_TABLE(1, (i * 4 + j));
		}
	}
	i = 0x76;		/* 6CT=> CHAN0 7CT=>CHAN1 */
	pmc->bar0[0x08] = i;
	/* Mapping to DMA buffers is the same as in case of acess point first come 0 framer and 0 time slot */
}

static void init_framer_regs(const struct quasar_pmc *pmc)
{

	int addr = 0x800, j;

	if (PMC_OLD_FIRMWARE == (pmc->bar0[0x0D] & 0xffff)) {
		pmc->bar0[0x03] = 0x00;	/* reset all framers */
		udelay(1000);
		pmc->bar0[0x03] = 0xff;
		udelay(1000);
	}

	pmc->bar1[addr] = 0x02;	/* setup page 2 */
	pmc->bar1[addr + 0x10] = 0xa0;	/* Configuration control word: E1 mode, Transmit enable */
	pmc->bar1[addr + 0x11] = 0;	/* P2.11=0h (Pulse attenuation) for 120 Om line. LIU control word */
	pmc->bar1[addr + 0x12] = 0;	/* write 0's to reserved space */
	pmc->bar1[addr + 0x14] = 0;
	pmc->bar1[addr + 0x15] = 0;
	pmc->bar1[addr + 0x16] = 0xbb;	/* P2.16 eqalizer high threshold */
	pmc->bar1[addr + 0x17] = 0x30;	/* eqalizer low threshold */

	pmc->bar1[addr] = 0x01;	/* setup page 1 */
	pmc->bar1[addr + 0x14] = 0x80;	/* to output data to dsto, select CAS */
	pmc->bar1[addr + 0x19] = 0x01;	/* to obtain loss signal */
	pmc->bar1[addr + 0x1f] = 0x18;	/* to setup automatic equaliser */
	pmc->bar1[addr + 0x15] = 0xa0;	/* P1.15 Coding and loopback controll HDB both directions */
	pmc->bar1[addr + 0x13] = 0x0b;	/* setup defaults for x bits? X123=111b */
	pmc->bar1[addr + 0x12] = 0x03;	/* setup defaults for International bits I01=11b */
	pmc->bar1[addr + 0x11] = 0x10;	/* disable TxAO */

	/* cas related */
	pmc->bar1[addr + 0x1e] = 0x1;	/* enable CAS interrupt */

	/* Enable to transmit CAS from registers on 5&6 pages */
	/* setup page 7 */
	pmc->bar1[addr] = 0x07;

	for (j = 0; j < 16; j++) {
		pmc->bar1[addr + 0x10 + j] = 0x02;

	}
	pmc->bar1[addr] = 0x08;
	for (j = 0; j < 16; j++) {
		pmc->bar1[addr + 0x10 + j] = 0x02;
	}
	/* fill 15ts CAS DATA nible with something to avoid starange initial shiftion of CAS, looks like framer hardware bug */
	pmc->bar1[addr] = 0x5;
	pmc->bar1[addr + 0x1f] = 0x5;
}

/*========== IOCTL (ecfg) ==========*/
/*
 static void check_bit(..)

 struct quasar_pmc * pmc - pointer to pmc device
 unsigned char chan_num - channel number
 unsigned char page - framer page 
 unsigned char line - framer register
 unsigned char smask - bit position in configuration uchar
 unsigned char hmask - bit position in hardware register
 unsigned char conf - data to write
 unsigned char *from_hw - readback
*/

static void check_bit(struct quasar_pmc *pmc,
		      unsigned char chan_num, unsigned char page,
		      unsigned char line, unsigned char smask,
		      const unsigned char hmask, unsigned char conf,
		      unsigned char *from_hw)
{
	unsigned int tmp;
	unsigned char sstore, htore, save_smask;

	tmp = framer_read(pmc, chan_num, page, line);
	htore = (unsigned char) (tmp & hmask);	/* prepare for comparison curent value from hw */
	save_smask = smask;
	sstore = conf & smask;	/* prepare for comparison value from config */
	if (smask > hmask) {	/* allign software value to hardware position */
		while (smask > hmask) {
			smask >>= 1;
			sstore >>= 1;
		}
	} else if (smask < hmask) {
		while (smask < hmask) {
			smask <<= 1;
			sstore <<= 1;
		}
	}
	if (htore ^ sstore) {	/* here sstore shifted to hardware position content of configuratuon value */
		tmp = framer_read(pmc, chan_num, page, line);
		tmp &= ~hmask;
		tmp |= (sstore & hmask);
		framer_write(pmc, chan_num, page, line, tmp);
	}
	/* reading back from hardware */
	tmp = framer_read(pmc, chan_num, page, line);
	sstore = (unsigned char) (tmp & hmask);
	smask = save_smask;
	if (smask > hmask) {	/* allign hardware to software position */
		while (smask > hmask) {
			smask >>= 1;
			sstore <<= 1;
		}
	} else if (smask < hmask) {
		while (smask < hmask) {
			smask <<= 1;
			sstore >>= 1;
		}
	}
	if (from_hw) {
		*from_hw |= sstore;
	}
}

/* Only allign to bit position in interface structure */
static void align_to_swposition(const unsigned int tmp,
				unsigned char smask,
				const unsigned char hmask,
				unsigned char *aligned)
{
	unsigned char sstore;
	sstore = (unsigned char) (tmp & hmask);
	if (smask > hmask) {	/* allign hardware to software position */
		while (smask > hmask) {
			smask >>= 1;
			sstore <<= 1;
		}
	} else if (smask < hmask) {
		while (smask < hmask) {
			smask <<= 1;
			sstore >>= 1;
		}
	}
	if (aligned) {
		*aligned |= sstore;
	}
}

static void framer2ecfg(struct quasar_pmc *pmc, unsigned char channo,
			unsigned long framer_reg, unsigned char ecfg_mask,
			unsigned char wcfg, unsigned char *rcfg)
{
	unsigned int tmp;

	if (channo & DIRECT_IN) {
		channo &= (~DIRECT_IN);
		tmp = framer_read(pmc, channo, PMC_GPAGE(framer_reg),
				PMC_GREG(framer_reg));
		align_to_swposition(tmp, ecfg_mask, PMC_GBIT(framer_reg), rcfg);
	} else {
		check_bit(pmc, channo, PMC_GPAGE(framer_reg),
			  PMC_GREG(framer_reg), ecfg_mask,
			  PMC_GBIT(framer_reg), wcfg, rcfg);
	}
}

/*
 Return codes:
 ENOTTY - invalid ioctl;
 EINVAL - passing invalid parameters;
 EFAULT - error copying data from userspace;
 0 - success;
*/
static int pmc_e1_ioctl(struct inode *inode_p, struct file *file_p,
			unsigned int cmd, unsigned long arg)
{
	int err = 0, k, j;
	unsigned int tmp = 0, channo = 0;
	struct pmc_config_chan wcfg, rcfg;
	struct pmc_err_ins err_ins;
	struct pmc_data_pass pass_data;
	struct pmc_status_chan stat;
	struct pmc_config_board bcfg;
	struct pmc_cas_xmit_data cas;
	struct pmc_version version;
	struct quasar_pmc *pmc =
	    (struct quasar_pmc *) file_p->private_data;

	if (_IOC_TYPE(cmd) != PTMC_E1_8_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > PTMC_E1_8_MCN)
		return -ENOTTY;
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *) arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void *) arg, _IOC_SIZE(cmd));
	if (err)
		return -EFAULT;

	memset((void *) &bcfg, 0, sizeof(struct pmc_config_board));
	memset((void *) &rcfg, 0, sizeof(struct pmc_config_chan));
	memset((void *) &wcfg, 0, sizeof(struct pmc_config_chan));
	memset((void *) &err_ins, 0, sizeof(struct pmc_err_ins));
	memset((void *) &stat, 0, sizeof(struct pmc_status_chan));
	memset((void *) &cas, 0, sizeof(struct pmc_cas_xmit_data));

	switch (cmd) {
	case PTMC_E1_CHAN_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CHAN_CONFIGURE.\n");
		if (copy_from_user
		    ((void *) &wcfg, (void *) arg, sizeof(wcfg))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		channo = wcfg.chan_num;
		if ((channo & (~DIRECT_IN)) > pmc->e1_amount)
			return -EINVAL;

		/* CFG0 */
		framer2ecfg(pmc, channo, CFG0_CAS_HP, CFG0_CAS, wcfg.cfg0,
			    &rcfg.cfg0);
		framer2ecfg(pmc, channo, CFG0_RCV_TR_HP, CFG0_RCV_TR,
			    wcfg.cfg0, &rcfg.cfg0);
		framer2ecfg(pmc, channo, CFG0_XMIT_TR_HP, CFG0_XMIT_TR,
			    wcfg.cfg0, &rcfg.cfg0);
		framer2ecfg(pmc, channo, CFG0_RCV_HDB3_HP, CFG0_RCV_HDB3,
			    wcfg.cfg0, &rcfg.cfg0);
		framer2ecfg(pmc, channo, CFG0_XMIT_HDB3_HP, CFG0_XMIT_HDB3,
			    wcfg.cfg0, &rcfg.cfg0);
		/* CFG1 */
		framer2ecfg(pmc, channo, CFG1_NAT0_HP, CFG1_NAT0,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pmc, channo, CFG1_NAT1_HP, CFG1_NAT1,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pmc, channo, CFG1_NAT2_HP, CFG1_NAT2,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pmc, channo, CFG1_NAT3_HP, CFG1_NAT3,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pmc, channo, CFG1_NAT4_HP, CFG1_NAT4,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pmc, channo, CFG1_INAT0_HP, CFG1_INAT0,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pmc, channo, CFG1_INAT1_HP, CFG1_INAT1,
			    wcfg.cfg1, &rcfg.cfg1);
		/* CFG2 */
		framer2ecfg(pmc, channo, CFG2_RAI_AUTO_ON_HP,
			    CFG2_RAI_MANUAL, wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pmc, channo, CFG2_RAI_ENA_HP, CFG2_RAI_ON,
			    wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pmc, channo, CFG2_RCAS_AUTO_ON_HP,
			    CFG2_RCAS_MANUAL, wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pmc, channo, CFG2_RCAS_ENA_HP, CFG2_RCAS_ON,
			    wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pmc, channo, CFG2_X1_HP, CFG2_X1, wcfg.cfg2,
			    &rcfg.cfg2);
		framer2ecfg(pmc, channo, CFG2_X2_HP, CFG2_X2, wcfg.cfg2,
			    &rcfg.cfg2);
		framer2ecfg(pmc, channo, CFG2_X3_HP, CFG2_X3, wcfg.cfg2,
			    &rcfg.cfg2);
		/* CFG3 */
		framer2ecfg(pmc, channo, CFG3_CRC4_TR_HP, CFG3_CRC4_TR,
			    wcfg.cfg3, &rcfg.cfg3);
		framer2ecfg(pmc, channo, CFG3_CRC4_DIS_HP, CFG3_CRC4_DIS,
			    wcfg.cfg3, &rcfg.cfg3);
		framer2ecfg(pmc, channo, CFG3_CRC4_MANUAL_HP,
			    CFG3_CRC4_MANUAL, wcfg.cfg3, &rcfg.cfg3);
		/* CFG4 */
		framer2ecfg(pmc, channo, CFG4_AIS_HP, CFG4_AIS, wcfg.cfg4,
			    &rcfg.cfg4);
		framer2ecfg(pmc, channo, CFG4_LLOOP_HP, CFG4_LLOOP,
			    wcfg.cfg4, &rcfg.cfg4);
		framer2ecfg(pmc, channo, CFG4_RLOOP_HP, CFG4_RLOOP,
			    wcfg.cfg4, &rcfg.cfg4);
		framer2ecfg(pmc, channo, CFG4_PLOOP_HP, CFG4_PLOOP,
			    wcfg.cfg4, &rcfg.cfg4);

		/* Write port configuration */
		if (!(channo & DIRECT_IN)) {
			/* configure board part */
			if (wcfg.cfg0 & CFG0_MASTER) {
				pmc->bar0[0x06] |= (unsigned int) (1 << channo);	/* restore bs/ls */
			} else {	/*allow to sync to this chan */
				pmc->bar0[0x06] &= (unsigned int) ~(1 << channo);	/* restore bs/ls */
			}
			tmp = pmc->bar0[0x04];	/* get TXAO */
			if (wcfg.cfg4 & CFG4_TX_DIS) {
				/* disable transmition, will transmit all ones */
				pmc->bar0[0x04] &= ~(1 << channo);	/* restore TXAO */
			} else {	/* enable transmition */
				pmc->bar0[0x04] |= (1 << channo);	/* restore TXAO */
			}
		}
		/* Read port configuration */
		channo &= (~DIRECT_IN);

		/* configure board part */
		/* get BS/LS */
		tmp = (unsigned char) pmc->bar0[0x06];
		align_to_swposition(tmp, CFG0_MASTER, 1 << channo, &rcfg.cfg0);	/* 0-6-chan_dep */


		/* get TXAO */
		tmp = ~(unsigned char) (pmc->bar0[0x04]);
		align_to_swposition(tmp, CFG4_TX_DIS, 1 << channo, &rcfg.cfg4);	/* 0-4-chan_dep */

		if (copy_to_user
		    ((void *) arg, (void *) &rcfg, sizeof(rcfg))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_ERR_EMU:
		module_dprintk(KERN_DEBUG, "PTMC_E1_ERR_EMU.\n");

		if (copy_from_user
		    ((void *) &err_ins, (void *) arg, sizeof(err_ins))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		channo = err_ins.chan_num;
		if (channo > pmc->e1_amount)
			return -EINVAL;

		/* generate errors */
		framer2ecfg(pmc, channo, ERR_EMU0_LOF_NFAS_HP,
			    ERR_EMU0_LOF_NFAS, err_ins.pmc_err, NULL);
		framer2ecfg(pmc, channo, ERR_EMU0_LOF_FAS_HP,
			    ERR_EMU0_LOF_FAS, err_ins.pmc_err, NULL);
		framer2ecfg(pmc, channo, ERR_EMU0_CRC4_HP, ERR_EMU0_CRC4,
			    err_ins.pmc_err, NULL);
		/* self restore zero state; */
		framer2ecfg(pmc, channo, ERR_EMU0_LOF_NFAS_HP,
			    ERR_EMU0_LOF_NFAS, 0, NULL);
		framer2ecfg(pmc, channo, ERR_EMU0_LOF_FAS_HP,
			    ERR_EMU0_LOF_FAS, 0, NULL);
		framer2ecfg(pmc, channo, ERR_EMU0_CRC4_HP, ERR_EMU0_CRC4,
			    0, NULL);

		break;
	case PTMC_E1_GET_STATS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_STATS.\n");

		if (copy_from_user
		    ((void *) &stat, (void *) arg, sizeof(stat))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		channo = stat.chan_num & (~DIRECT_IN);
		if (channo > pmc->e1_amount)
			return -EINVAL;

		framer2ecfg(pmc, channo, STAT0_LOS_HP, STAT0_LOS, 0,
			    &stat.stat0);
		framer2ecfg(pmc, channo, STAT0_LOF_HP, STAT0_LOF, 0,
			    &stat.stat0);
		framer2ecfg(pmc, channo, STAT0_CAS_LOM_HP, STAT0_CAS_LOM,
			    0, &stat.stat0);
		framer2ecfg(pmc, channo, STAT0_CRC4_SYNC_HP,
			    STAT0_CRC4_SYNC, 0, &stat.stat0);
		framer2ecfg(pmc, channo, STAT0_AIS_HP, STAT0_AIS, 0,
			    &stat.stat0);
		framer2ecfg(pmc, channo, STAT0_RAIS_HP, STAT0_RAIS, 0,
			    &stat.stat0);
		tmp = framer_read(pmc, channo, 0x3, 0x10);
		if (!((tmp & PMC_GBIT(STAT0_E_BIT1_HP))
		     && (tmp & PMC_GBIT(STAT0_E_BIT2_HP))))
			stat.stat0 |= STAT0_E_BIT;

		tmp = framer_read(pmc, channo, 0x3, 0x11);
		if ((tmp & PMC_GBIT(STAT0_CRC4_1_HP))
		    || (tmp & PMC_GBIT(STAT0_CRC4_2_HP)))
			stat.stat0 |= STAT0_CRC4;

		framer2ecfg(pmc, channo, STAT1_NAT0_HP, STAT1_NAT0, 0,
			    &stat.stat1);
		framer2ecfg(pmc, channo, STAT1_NAT1_HP, STAT1_NAT1, 0,
			    &stat.stat1);
		framer2ecfg(pmc, channo, STAT1_NAT2_HP, STAT1_NAT2, 0,
			    &stat.stat1);
		framer2ecfg(pmc, channo, STAT1_NAT3_HP, STAT1_NAT3, 0,
			    &stat.stat1);
		framer2ecfg(pmc, channo, STAT1_NAT4_HP, STAT1_NAT4, 0,
			    &stat.stat1);
		framer2ecfg(pmc, channo, STAT1_INAT0_HP, STAT1_INAT0, 0,
			    &stat.stat1);
		framer2ecfg(pmc, channo, STAT1_INAT1_HP, STAT1_INAT1, 0,
			    &stat.stat1);
		framer2ecfg(pmc, channo, STAT1_SLIP_HP, STAT1_SLIP, 0,
			    &stat.stat1);

		framer2ecfg(pmc, channo, STAT2_Y_HP, STAT2_Y, 0,
			    &stat.stat2);
		framer2ecfg(pmc, channo, STAT2_X1_HP, STAT2_X1, 0,
			    &stat.stat2);
		framer2ecfg(pmc, channo, STAT2_X2_HP, STAT2_X2, 0,
			    &stat.stat2);
		framer2ecfg(pmc, channo, STAT2_X3_HP, STAT2_X3, 0,
			    &stat.stat2);

		framer2ecfg(pmc, channo, OVR_HDB3_HP, OVR_HDB3, 0,
			    &stat.ovr);
		framer2ecfg(pmc, channo, OVR_FAS_HP, OVR_FAS, 0,
			    &stat.ovr);
		framer2ecfg(pmc, channo, OVR_CRC4_HP, OVR_CRC4, 0,
			    &stat.ovr);

		/* reading FAS counter */
		stat.fas = framer_read_pr(pmc, channo, CNT_FAS_HP);
		/* reading HDB3 counter */
		stat.hdb3 = framer_read_pr(pmc, channo, CNT_BPVHI_HP);
		stat.hdb3 <<= 8;
		stat.hdb3 |= framer_read_pr(pmc, channo, CNT_BPVLO_HP);
		/* reading CRC4 counter */
		stat.crc4 = framer_read_pr(pmc, channo, CNT_CRC4HI_HP);
		stat.crc4 <<= 8;
		stat.crc4 |= framer_read_pr(pmc, channo, CNT_CRC4LO_HP);

		/* clear counter if requested */
		if (stat.mgmt & OVR_FAS) {
			framer_write_pr(pmc, channo, CNT_FAS_HP, 0);
		}
		if (stat.mgmt & OVR_HDB3) {
			framer_write_pr(pmc, channo, CNT_BPVHI_HP, 0);
			framer_write_pr(pmc, channo, CNT_BPVLO_HP, 0);
		}
		if (stat.mgmt & OVR_CRC4) {
			framer_write_pr(pmc, channo, CNT_CRC4HI_HP, 0);
			framer_write_pr(pmc, channo, CNT_CRC4LO_HP, 0);
		}

		/* get cas signalling */
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pmc, channo, 0x9, 0x10 + j);
			if (j < (2 * k + 1)) {
				stat.cas_data[k] =
				    (unsigned char) (tmp & 0xf);
				stat.cas_data[k] <<= 4;

			} else {
				tmp &= 0x0f;
				stat.cas_data[k] |= (unsigned char) tmp;
				k += 1;
			}
		}
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pmc, channo, 0xA, 0x10 + j);
			if (j < (2 * k + 1)) {
				stat.cas_data[8 + k] =
				    (unsigned char) (tmp & 0xf);
				stat.cas_data[8 + k] <<= 4;
			} else {
				tmp &= 0x0f;
				stat.cas_data[8 + k] |=
				    (unsigned char) tmp;
				k += 1;
			}
		}

		if (copy_to_user
		    ((void *) arg, (void *) &stat, sizeof(stat))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_RESET:
		/* Reset board. Not tested very well. */
		module_dprintk(KERN_DEBUG, "PTMC_E1_RESET.\n");

		/* initialising board registers */
		init_ptmc_regs(pmc);
		/* initialising framer registers */
		init_framer_regs(pmc);
		/* initialising time-slots maping table */
		init_switch_map(pmc);
		break;
	case PTMC_E1_BOARD_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_BOARD_CONFIGURE.\n");

		if (copy_from_user((void *) &bcfg, (void *) arg, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		/* reading ctrl register */
		tmp = pmc->bar0[0x02];
		tmp &= ~(MODE1_NF1_GEN_ENA_HP | MODE1_NF2_GEN_ENA_HP \
			 | MODE1_NF1_USE_ENA_HP | MODE1_NF2_USE_ENA_HP \
			 | MODE1_AB_HP | MODE1_MA_SL_SEL_HP);
		if (bcfg.mode1 & MODE1_NF1_GEN_ENA) {
			tmp |= MODE1_NF1_GEN_ENA_HP;
		}
		if (bcfg.mode1 & MODE1_NF2_GEN_ENA) {
			tmp |= MODE1_NF2_GEN_ENA_HP;
		}
		if (bcfg.mode1 & MODE1_NF1_USE_ENA) {
			tmp |= MODE1_NF1_USE_ENA_HP;
		}
		if (bcfg.mode1 & MODE1_NF2_USE_ENA) {
			tmp |= MODE1_NF2_USE_ENA_HP;
		}
		if (bcfg.mode1 & MODE1_AB) {
			tmp |= MODE1_AB_HP;
		}
		if (bcfg.mode1 & MODE1_MA_SL_SEL) {
			tmp |= MODE1_MA_SL_SEL_HP;
		}
		pmc->bar0[0x02] = tmp;
		udelay(2000);
		break;
	case PTMC_E1_GET_PCI_SLOT:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_SLOT.\n");

		tmp = PCI_SLOT((unsigned int) pmc->pci_device->devfn);
		if (copy_to_user((void *)arg, (void *)&tmp, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_PCI_BUSNUM:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_BUSNUM.\n");

		tmp = pmc->pci_device->bus->number;
		if (copy_to_user((void *)arg, (void *)&tmp, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_READ_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR0.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pass_data.rdata = pmc->bar0[pass_data.raddress];
		if (copy_to_user((void *)arg, (void *)&pass_data, sizeof(pass_data))) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR0.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pmc->bar0[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_READ_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR1.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pass_data.rdata = pmc->bar1[pass_data.raddress];
		if (copy_to_user((void *)arg, (void *)&pass_data, sizeof(pass_data))) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR1.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pmc->bar1[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_CAS_XMIT_PASS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CAS_XMIT_PASS\n");

		if (copy_from_user((void *)&cas, (void *)arg, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		channo = cas.chan_num;
		if ((channo & (~DIRECT_IN)) > pmc->e1_amount)
			return -EINVAL;

		if (!(channo & DIRECT_IN)) {	/* writing cas data */
			channo &= (~DIRECT_IN);
			module_dprintk(KERN_DEBUG,
				       "Pass data! Chan_num = %d\n",
				       channo);

			k = 0;
			for (j = 0; j < 16; j++) {
				if (j < (2 * k + 1)) {
					tmp = (cas.cdata[k] >> 4) & 0x0f;
					framer_write(pmc, channo, 0x5,
						     0x10 + j, tmp);
				} else {
					tmp = (cas.cdata[k] & 0x0f);
					framer_write(pmc, channo, 0x5,
						     0x10 + j, tmp);
					k += 1;
				}
			}
			k = 0;
			for (j = 0; j < 16; j++) {
				if (j < (2 * k + 1)) {
					tmp =
					    (cas.cdata[8 + k] >> 4) & 0x0f;
					framer_write(pmc, channo, 0x6,
						     0x10 + j, tmp);
				} else {
					tmp = (cas.cdata[8 + k] & 0x0f);
					framer_write(pmc, channo, 0x6,
						     0x10 + j, tmp);
					k += 1;
				}
			}
			memset((void *) &cas, 0, sizeof(struct pmc_cas_xmit_data));	/* clear to get values back */
		}
		/* Reading back cas signalling */
		channo &= (~DIRECT_IN);
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pmc, channo, 0x5, 0x10 + j);
			if (j < (2 * k + 1)) {
				tmp &= 0x0f;
				cas.cdata[k] = (unsigned char) tmp;
				cas.cdata[k] <<= 4;
			} else {
				tmp &= 0x0f;
				cas.cdata[k] |= (unsigned char) tmp;
				k += 1;
			}

		}
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pmc, channo, 0x6, 0x10 + j);
			if (j < (2 * k + 1)) {
				tmp &= 0x0f;
				cas.cdata[8 + k] = (unsigned char) tmp;
				cas.cdata[8 + k] <<= 4;
			} else {
				tmp &= 0x0f;
				cas.cdata[8 + k] |= (unsigned char) tmp;
				k += 1;
			}
		}

		if (copy_to_user((void *) arg, (void *) &cas, sizeof(cas))) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_VERSION:
		module_dprintk(KERN_DEBUG, "PTMC_E1_VERSION.\n");
		strncpy(version.proto_version, ecfg_proto_version,
			sizeof(version.proto_version));
		version.hw_fw_rev = pmc->bar0[0x0D];
		strncpy(version.driver_version, QUASAR_VERSION,
			sizeof(version.driver_version));
		if (copy_to_user
		    ((void *) arg, (void *) &version, sizeof(version))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
		/*
		   case PTMC_E1_PASS_ABCD_ZAP_FAKE: removed.
		 */
	default:
		break;
	}
	return 0;
}

#ifdef HAVE_UNLOCKED_IOCTL
#  if !defined(CONFIG_BKL)
#    define lock_kernel();
#    define unlock_kernel();
#  endif //CONFIG_BKL
static long pmc_e1_ioctl_unlocked(struct file *file, unsigned int cmd,
				  unsigned long data)
{
	long res;

	lock_kernel();
	res = pmc_e1_ioctl(NULL, file, cmd, data);
	unlock_kernel();
	return res;
}
#endif //HAVE_UNLOCKED_IOCTL

static int pmc_e1_open(struct inode *inode_p, struct file *file_p)
{
	file_p->private_data = container_of(inode_p->i_cdev, struct quasar_pmc, dev_file);
	return 0;
}

static int pmc_e1_release(struct inode *inode_p, struct file *file_p)
{
	return 0;
}

static struct file_operations pmc_e1_fops = {
      owner:THIS_MODULE,
      open:pmc_e1_open,
#ifdef HAVE_UNLOCKED_IOCTL
      unlocked_ioctl:pmc_e1_ioctl_unlocked,
#else
      ioctl:pmc_e1_ioctl,
#endif
      release:pmc_e1_release,
};

/*========== SYSFS code ==========*/
#ifdef PMC_E1_UDEV
struct ptmc_attribute {
	struct device_attribute attr;
	unsigned long p[3];
};
#define toptattr(a) ((struct ptmc_attribute *)(a))
#define todeattr(a) ((struct device_attribute *)(a))

#define PTMC_ATTR_RO(name) __ATTR(name, S_IRUGO, _pmc_show, NULL)
ssize_t _pmc_show(struct device *cls, struct device_attribute *attr,
		  char *buff)
{
	struct quasar_pmc *pmc;
	unsigned long *p = toptattr(attr)->p;

	pmc = (struct quasar_pmc *) dev_get_drvdata(cls);
	return sprintf(buff, "%d\n", (pmc->bar0[p[0]] >> p[1]) && p[2]);
}

ssize_t pmc_show(struct device * cls, struct device_attribute * attr,
		 char *buff)
{
	int count = 0, i, val;
	unsigned int tmp;
	struct quasar_pmc *pmc;
	pmc = (struct quasar_pmc *) dev_get_drvdata(cls);

	count += sprintf(buff, "ChN:");
	for (i = 0; i <= pmc->e1_amount; i++)
		count += sprintf(buff + count, " %d", i);

	count += sprintf(buff + count, "\nLOS:");
	tmp = pmc->bar0[0x05];
	for (i = 0; i <= pmc->e1_amount; i++) {
		if (tmp & (1 << i))
			val = 1;
		else
			val = 0;
		count += sprintf(buff + count, " %d", val);
	}

	count += sprintf(buff + count, "\nLOF:");
	for (i = 0; i <= pmc->e1_amount; i++) {
		if (tmp & framer_checkbit(pmc, i, STAT0_LOF_HP))
			val = 1;
		else
			val = 0;
		count += sprintf(buff + count, " %d", val);
	}
	count += sprintf(buff + count, "\nLOM:");
	for (i = 0; i <= pmc->e1_amount; i++) {
		if (tmp & framer_checkbit(pmc, i, STAT0_CAS_LOM_HP))
			val = 1;
		else
			val = 0;
		count += sprintf(buff + count, " %d", val);
	}

	count += sprintf(buff + count, "\n");
	return count;
}

static struct ptmc_attribute device_attrs[] = {
	{PTMC_ATTR_RO(pmc_load), {0x21, 16, 0xFF}},
	{PTMC_ATTR_RO(pmc_load_max), {0x21, 0, 0xFF}},
	{PTMC_ATTR_RO(pci_load), {0x20, 16, 0xFF}},
	{PTMC_ATTR_RO(pci_load_cur), {0x20, 0, 0xFF}},
	{PTMC_ATTR_RO(dma_overflow), {0x09, 24, 0xFF}},
	{__ATTR(info, S_IRUGO, pmc_show, NULL), {0,}},
};

static int ptmc_sysfs_create(struct quasar_pmc *pmc)
{
	int i, error = 0;
	struct device *ptmc_device;

	pmc->pmc_class_dev = ptmc_device =
		QUASAR_DEVICE_CREATE(pmc_e1_class, QUASAR_PARENT(pmc),
				     MKDEV(MAJOR(dev_file_glob),
					   pmc->dev_number), pmc->name);
	if (IS_ERR(pmc->pmc_class_dev)) {
		module_printk(KERN_WARNING, "Failed to device_create.\n");
		return 1;
	}
	dev_set_drvdata(pmc->pmc_class_dev, (void *) pmc);

	for (i = 0; i < ARRAY_SIZE(device_attrs); i++) {
		error = device_create_file(ptmc_device, &(device_attrs[i].attr));
		if (error)
			break;
	}

	if (error) {
		while (--i >= 0)
			device_remove_file(ptmc_device,
					   todeattr(&device_attrs[i]));
	}

	return 0;
}

static void ptmc_sysfs_delete(struct quasar_pmc *pmc)
{
	int i;
	struct device *ptmc_device = pmc->pmc_class_dev;

	if (IS_ERR(pmc->pmc_class_dev))
		return;

	for (i = 0; i < ARRAY_SIZE(device_attrs); i++)
		device_remove_file(ptmc_device,
				   todeattr(&device_attrs[i]));

	device_destroy(pmc_e1_class,
		       MKDEV(MAJOR(dev_file_glob), pmc->dev_number));
}
#endif				/* PMC_E1_UDEV */

/*========== Init code & IRQ handler ==========*/
QUASAR_IRQ_HANDLER(ptmc_e1_interrupt_heandler)
{
	unsigned int flags = 0;
	struct quasar_pmc *pmc = (struct quasar_pmc *) pr_data;
	if (pmc == NULL)
		return IRQ_NONE;
	flags = pmc->bar0[0];
	if (flags & 0x20000000) {	/* check DMA interrupt; */
		pmc->bar0[0x0] = 0x0;
		e12zap_delivery(pmc);
		return IRQ_HANDLED;
	}
	return IRQ_NONE;
}

static int set_cacheline(struct pci_dev *pdev, u8 val)
{
	u8 cache_tmp;

	/* Write the correct value. */
	pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, val);
	/* Read it back. */
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &cache_tmp);
	if (val != cache_tmp) {
		module_printk(KERN_ERR,
			      "Cache line size of %d DWORDs is not "
			      "supported (%x:%x)\n", val, pdev->vendor,
			      pdev->device);
		return -EINVAL;
	}

	return 0;
}

/*
 Some BIOS or Linux, set PCIe bridge "cache line size" to illegal value (>20).
 Setting "cache line size" for our FPGA & PCIe bridge (PEX 8111, PEX 8112).
 Is PCIe bridge the only device on the PCI bus (can I use first device) ?
*/
int tune_pci(struct pci_dev *pdev)
{
	struct pci_dev *bridge;
	u8 lat_timer, tmp_u8;

	module_printk(KERN_INFO, "Setting Cache Line Size to %d\n", cacheline);
	if (set_cacheline(pdev, cacheline))
		return -EINVAL;

	pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &lat_timer);
	if (lat_timer < 40 || latency) {
		lat_timer = latency ? latency : 40;
		module_printk(KERN_INFO, "Setting latency timer to %d\n", lat_timer);
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER, lat_timer);
		pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &tmp_u8);
		if (tmp_u8 != lat_timer) {
			printk("%d %d\n", tmp_u8, lat_timer);
			module_printk(KERN_ERR, "Latency timer of %d is not "
				      "supported (%x:%x)\n", lat_timer,
				      pdev->vendor, pdev->device);
			return -EINVAL;
		}
	}

	/* Configuring PCIe bridge (PEX), if any. */
	if (!pdev->bus || !pdev->bus->self) {
		module_printk (KERN_ERR, "tune_pci(): Failed to locate bridge.\n");
		return 0;
	}
        bridge = pdev->bus->self;
	if (!bridge || (0x10b5 != bridge->vendor) ||
	    (0x8111 != bridge->device && 0x8112 != bridge->device))
		return 0;

	module_printk(KERN_INFO, "Setting PCIe bridge Cache Line Size to %d\n", cacheline_bridge);
	if (set_cacheline(bridge, cacheline_bridge))
		return -EINVAL;

	return 0;
}

/*
Allocates memory for pmc_e1_controller structure, returns a pointer
to allocated strunture, NULL if allocation insucessful.	*/
static struct quasar_pmc *pmc_e1_alloc(void)
{
	struct quasar_pmc *pmc;
	/* create structure for pmc_device */
	pmc = kmalloc(sizeof(*pmc), GFP_KERNEL);
	if (pmc) {
		memset(pmc, 0, sizeof(*pmc));
	}
	return pmc;
}

static int ptmc_e1_init_one(struct pci_dev *pdev,
			    const struct pci_device_id *ent)
{
	struct quasar_pmc *pmc;
	int err = 0;
	u8 lat_timer;

	if (pci_enable_device(pdev))
		return -EIO;
	if (device_number > (PMC_E18_NR - 1))
		return -EBUSY;

	err = tune_pci(pdev);
	if (err)
		return err;

	pmc = pmc_e1_alloc();
	if (!pmc) {
		module_printk(KERN_ERR, "Failed to allocate device.\n");
		return -ENOMEM;
	}
	/* TODO: move to prepare_io */
	pmc->dev_number = device_number;
	sprintf(pmc->name, "quasar%d", device_number);
	pci_read_config_byte(pdev, PCI_REVISION_ID, &lat_timer);
	switch (lat_timer) {
	case 0xa2:
		pmc->e1_amount = 1;
		break;
	case 0xa4:
		pmc->e1_amount = 3;
		break;
	case 0xa8:
		pmc->e1_amount = 7;	/* [0:7] */
		break;
	default:
		pmc->e1_amount = 7;	/* assume old revision device 8 ports only. */
		break;
	}

	/* bind with pci device; */
	pmc->pci_device = pdev;
	/* add to list of controllers, I don't remember why should I use this list, just part of old code from old project. */
	list_add_tail(&pmc->dev_node, &dev_list_head);
	/*  printk("pciaddr = 0x%lx, irq = %d, ptmc_dev = %p \n", pci_resource_start(pdev,0), pdev->irq, pmc); */
	if (request_mem_region(pci_resource_start(pdev, 0), WIN_SIZE, pmc->name) == NULL) {
		module_printk(KERN_ERR,
			      "I/O resource 0x%x @ 0x%llx busy\n",
			      WIN_SIZE, (u64) pci_resource_start(pdev, 0));
		goto fail_req_bar0;
	}
	if (request_mem_region(pci_resource_start(pdev, 1), WIN_SIZE, pmc->name) == NULL) {
		module_printk(KERN_ERR,
			      "I/O resource 0x%x @ 0x%llx busy\n",
			      WIN_SIZE, (u64) pci_resource_start(pdev, 1));
		goto fail_req_bar1;
	}
	pmc->bar0 = ioremap_nocache(pci_resource_start(pdev, 0), WIN_SIZE);
	if (!pmc->bar0) {
		module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
			      WIN_SIZE, (u64) pci_resource_start(pdev, 0));
		goto fail_map_bar0;
	}
	pmc->bar1 = ioremap_nocache(pci_resource_start(pdev, 1), WIN_SIZE);
	if (!pmc->bar1) {
		module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
			      WIN_SIZE, (u64) pci_resource_start(pdev, 1));
		goto fail_map_bar1;
	}
	ptmc_sysfs_create(pmc);

	module_printk(KERN_INFO,
		      "Registering interrupt %d with flags 0x%x\n",
		      pdev->irq, irqflags);
	if (request_irq(pdev->irq, ptmc_e1_interrupt_heandler, irqflags, pmc->name,
	     (void *) pmc) < 0) {
		module_printk(KERN_ERR, "Can't register interrupt: %d\n", pdev->irq);
		goto fail_reg_irq;
	}
	cdev_init(&pmc->dev_file, &pmc_e1_fops);
	pmc->dev_file.owner = THIS_MODULE;
	pmc->dev_file.ops = &pmc_e1_fops;
	if (cdev_add(&pmc->dev_file, MKDEV(MAJOR(dev_file_glob), device_number), 1)) {
		module_printk(KERN_ERR, "Can't register device file for: %d\n", device_number);
		goto fail_req_file;
	}
	device_number++;	/* for non-hot plug systems, for hot-plug system we probably should change system of numbering... */
	spin_lock_init(&pmc->dev_lock);
	/* mark module as initialised */
	pmc->registered = 1;

	/* Reset and initialise board; */

	/* initialising board registers */
	init_ptmc_regs(pmc);

	/* init framer registers */
	init_framer_regs(pmc);

	/* initialising time-slots maping table. */
	init_switch_map(pmc);

	/* DAHDI */
	module_printk(KERN_INFO, "Quasar board with %d ports found, dev id: %d, hw_fw_rev: 0x%x, (pci slot %d)\n",
		      pmc->e1_amount + 1, pmc->dev_number, pmc->bar0[0x0D], PCI_SLOT(pdev->devfn) + 1);
	err = e12zap_register(pmc);
	if (err)
		module_printk(KERN_ERR, "Quasar dahdi registration error.\n");

	return 0;
	/* Clean up all resources, but leave in list. */
fail_req_file:
	free_irq(pdev->irq, (void *) pmc);
fail_reg_irq:
	ptmc_sysfs_delete(pmc);
	iounmap((void *) pmc->bar1);
fail_map_bar1:
	iounmap((void *) pmc->bar0);
fail_map_bar0:
	release_mem_region(pci_resource_start(pdev, 1), WIN_SIZE);
/* unreg 1 bar; */
fail_req_bar1:
	release_mem_region(pci_resource_start(pdev, 0), WIN_SIZE);
/* unreg 0 bar; */
fail_req_bar0:
	pmc->registered = 0;
	return -EBUSY;
}

static void ptmc_e1_remove_one(struct pci_dev *pdev)
{
	struct list_head *pos, *n;
	struct quasar_pmc *pmc = NULL;
	list_for_each_safe(pos, n, &dev_list_head) {
		if (list_entry(pos, struct quasar_pmc, dev_node)->
		    pci_device == pdev) {
			pmc = list_entry(pos, struct quasar_pmc, dev_node);
		}
	}
	if (pmc) {
		if (pmc->registered) {
			e12zap_unregister(pmc);	/* DAHDI */
			free_irq(pdev->irq, (void *) pmc);
			cdev_del(&pmc->dev_file);
			ptmc_sysfs_delete(pmc);

			iounmap((void *) pmc->bar0);
			release_mem_region(pci_resource_start(pdev, 0),
					   WIN_SIZE);
			iounmap((void *) pmc->bar1);
			release_mem_region(pci_resource_start(pdev, 1),
					   WIN_SIZE);

		}
		/* unlink from device's list */
		list_del(&pmc->dev_node);
		kfree(pmc);
	}
}

static struct pci_driver ptmc_e1_driver = {
      name:DRV_NAME,
      probe:ptmc_e1_init_one,
      remove:ptmc_e1_remove_one,
      id_table:ptmc_e1_tbl,
};

static int __init ptmc_e1_init_module(void)
{
	int result, error;
	module_printk(KERN_INFO,
		      "Quasar E1 interface board driver, Parabel Ltd. http://parabel.ru, "
		      "Version: %s\n", QUASAR_VERSION);

	if (PMC_E1_MAJOR) {
		dev_file_glob = MKDEV(PMC_E1_MAJOR, 0);
		result = register_chrdev_region(dev_file_glob, PMC_E18_NR, DRV_NAME);
	} else {
		result = alloc_chrdev_region(&dev_file_glob, 0, PMC_E18_NR, DRV_NAME);
	}
	if (result < 0) {
		module_printk(KERN_ERR, "Can't register major number for Quasar driver.\n");
		return -EBUSY;
	}
#ifdef PMC_E1_UDEV
	pmc_e1_class = class_create(THIS_MODULE, DRV_NAME);
	if (IS_ERR(pmc_e1_class)) {
		module_printk(KERN_ERR, "Can't create class " DRV_NAME ".\n");
		error = -PTR_ERR(pmc_e1_class);
		goto fail_reg_class;
	}
#endif
	result = pci_register_driver(&ptmc_e1_driver);
	if (result >= 0) {
		return 0;
	} else {
		module_printk(KERN_ERR, "PCI registration for Quasar board failed!");
		error = -EBUSY;
	}
	pci_unregister_driver(&ptmc_e1_driver);

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
      fail_reg_class:
#endif


	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	return error;
}

static void __exit ptmc_e1_cleanup_module(void)
{
	pci_unregister_driver(&ptmc_e1_driver);
	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	module_printk(KERN_INFO, "Quasar E1 interface board driver unloaded.\n");

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
#endif

	return;
}

module_init(ptmc_e1_init_module);
module_exit(ptmc_e1_cleanup_module);

module_param(cacheline, int, 0600);
MODULE_PARM_DESC(cacheline, "PCI Cache Line Size (in units of DWORDs).");
module_param(cacheline_bridge, int, 0600);
MODULE_PARM_DESC(cacheline_bridge, "PCIe bridge Cache Line Size (DWORDs).");
module_param(latency, int, 0400);
MODULE_PARM_DESC(latency, "PCI BUS latency timer.");
module_param(irqflags, int, 0600);
MODULE_PARM_DESC(irqflags, "Interrupt handler flags (passed to request_irq).");

MODULE_AUTHOR("Khelik Mikhail <mixel.net@gmail.com>, Pavel Selivanov <pavel[AT]parabel.ru>");
MODULE_DESCRIPTION("Quasar E1 interface board driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(QUASAR_VERSION);
